#include "header\gba.h"
#include "header\bg.h"
#include "header\timer.h"
#include "header\dma.h"
#include "header\irq.h"
#include "header\sound.h"
#include "header\sprite.h"
#include "header\blend.h"
#include "header\mathtable.h"

#include "lib.h"
#include "syscall.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <malloc.h>

ST_FONTBITMAP FONTBITMAP;
ST_FONTSJIS   FONTSJIS;
ST_WAVE       WAVE;
ST_KEY        KEY;
ST_SEEK       SEEK;
ST_DMA_CHAIN  DMA_CHAIN[MAX_DMA_CHAIN];
ST_TIMER      Timer;
ST_INPUT      Input;

//ROMt@C̏I[|C^
extern u8 __exidx_start;

//========================================================================
void WaitForVsync(void)
{
	while(REG_VCOUNT != 159){};
	while(REG_VCOUNT == 159){};
}
//========================================================================
u16 Mode3GetDotRGB(u8 x,u8 y)
{
	u16* Screen = VideoBuffer;
	return Screen[y*240+x];
}
//========================================================================
void Mode3DrawDot(u8 x,u8 y,u16 color)
{
	u16* Screen = VideoBuffer;
	Screen[y*240+x] = color;
}
//========================================================================
void Mode3DrawLine(u8 x1,u8 y1,u8 x2,u8 y2,u16 color)
{
	s16 E,x,y;
	s16 dx,dy,sx,sy,i;

	sx = ( x2 > x1 ) ? 1 : -1;
	dx = ( x2 > x1 ) ? x2 - x1 : x1 - x2;
	sy = ( y2 > y1 ) ? 1 : -1;
	dy = ( y2 > y1 ) ? y2 - y1 : y1 - y2;

	x = x1;
	y = y1;

	if(dx >= dy)
	{	//X1ȉ̏ꍇ
		E = -dx;

		for(i=0; i<=dx; i++)
		{
			Mode3DrawDot(x, y, color);
			x += sx;
			E += 2 * dy;
			if(E >= 0)
			{
				y += sy;
				E -= 2 * dx;
			}
		}
	}
	else
	{	//X1傫ꍇ
		E = -dy;
		
		for(i=0; i<=dy; i++)
		{
			Mode3DrawDot(x, y, color);
			y += sy;
			E += 2 * dx;
			if(E >= 0)
			{
				x += sx;
				E -= 2 * dy;
			}
		}
	}
}
//========================================================================
void Mode3DrawBox(u8 x1,u8 y1,u8 x2,u8 y2,u16 color)
{
	u16* Screen = VideoBuffer;
	u8 x,y;

	for(y=y1; y<y2; y++)
	{
		for(x=x1; x<x2; x++)
		{
			Screen[y*240+x] = color;
		}
	}
}
//========================================================================
void Mode3DrawCls(u16 color)
{
	Mode3DrawBox(0,0,240,160,color);
}
//========================================================================
void Mode3DrawImage(u8 x1, u8 y1, u8 cx, u8 cy, u16* image)
{
	u16* Screen = VideoBuffer;
	u8 x,y;

	Screen += y1*240 + x1;

	for(y=0; y<cy; y++)
	{
		for(x=0; x<cx; x++)
		{
			Screen[y*240+x] = image[y*240+x];
		}
	}
}
//========================================================================
void DMA3MemcpyFIX16(u32 src, u32 dest, u16 size)
{
	REG_DM3SAD   = src;
	REG_DM3DAD   = dest;
	REG_DM3CNT_L = size;
	REG_DM3CNT_H = (DMA_SIZE_16    | DMA_TRANSFER_ON |
	                DMA_TIMING_NOW | DMA_REPEAT_OFF  |
	                DMA_SAD_FIX    | DMA_DAD_INC     | DMA_INTR_OFF);

	while(REG_IF&IRQ_BIT_DMA3){}    //DMA]tO~܂Ń[v
}
//========================================================================
void DMA3MemcpyINC16(u32 src, u32 dest, u16 size)
{
	REG_DM3SAD   = src;
	REG_DM3DAD   = dest;
	REG_DM3CNT_L = size;
	REG_DM3CNT_H = (DMA_SIZE_16    | DMA_TRANSFER_ON |
	                DMA_TIMING_NOW | DMA_REPEAT_OFF  |
	                DMA_SAD_INC    | DMA_DAD_INC     | DMA_INTR_OFF);

	while(REG_IF&IRQ_BIT_DMA3){}    //DMA]tO~܂Ń[v
}
//========================================================================
void DMA3MemcpyINC32(u32 src, u32 dest, u16 size)
{
	REG_DM3SAD   = src;
	REG_DM3DAD   = dest;
	REG_DM3CNT_L = size;
	REG_DM3CNT_H = (DMA_SIZE_32    | DMA_TRANSFER_ON |
	                DMA_TIMING_NOW | DMA_REPEAT_OFF  |
	                DMA_SAD_INC    | DMA_DAD_INC     | DMA_INTR_OFF);

	while(REG_IF&IRQ_BIT_DMA3){}    //DMA]tO~܂Ń[v
}
//========================================================================
void Mode3SetBitMapFont(u8 *offset, u16 cx, u16 cy)
{
	FONTBITMAP.offset = (u16*)offset;
	FONTBITMAP.cx     = cx;				//1Font̑傫
	FONTBITMAP.cy     = cy;
	FONTBITMAP.size   = cx*cy;
}
//========================================================================
void Mode3SetSJISFont(u8* sheet, u8* offset, u16 cnt, u8 cx, u8 cy)
{
	FONTSJIS.FontSheet = (u16*)sheet;	//tHgV[g
	FONTSJIS.Offset    = (u8 *)offset;	//tHgf[^
	FONTSJIS.cx        = cx;
	FONTSJIS.cy        = cy;
	FONTSJIS.size      = cx*cy;
	FONTSJIS.cnt       = cnt/2;         //tHg
}
//========================================================================
void Mode3DrawBitMapChar(u8 x,u8 y,u16 chr)
{
	u16 *Screen = VideoBuffer;
	u16 *Offset = FONTBITMAP.offset;
	u8 i,j;

	Screen += FONTBITMAP.cx*x + FONTBITMAP.cy*y*240;
	Offset += FONTBITMAP.size * chr;

	for(i=0; i<FONTBITMAP.cy; i++)
	{
		for(j=0; j<FONTBITMAP.cx; j++)
		{
			Screen[i*240+j] = *Offset++;
		}
	}
}
//========================================================================
void Mode3DrawBitMapStr(u8 x,u8 y, u8* str)
{
	u16 c;
	u8 i=0,dx=0;

	while(str[i]!=0x00 && str[i]!=0x0a)
	{
		if(str[i]<=0x7f)	//1byte char
		{
			Mode3DrawBitMapChar(x+dx,y,(u16)str[i]);
			dx++; i++;
		}
		else
		{					//2byte char
			switch(str[i])
			{
				case 0x82: c = str[i+1]-0x90+192; break;
				case 0x83: c = str[i+1]-0x40+304; break;
				default  : c = str[i+1];          break;
			}
			Mode3DrawBitMapChar(x+dx,y,c);
			dx++; i+=2;
		}
	}
}
//========================================================================
void BGSetChrData(u8 No,u16 *img,u32 size)
{
	u16 *BGChr = MEM_BG_CHR(No);
	u32 i;

	for(i=0; i<size/2; i++)
	{
		*BGChr++ = *img++;
	}
}
//========================================================================
void BGSetMapData(u8 No,u16 *img,u32 size)
{
	u16 *BGMap = MEM_BG_MAP(No);
	u32 i;

	for(i=0; i<size/2; i++)
	{
		*BGMap++ = *img++;
	}
}
//========================================================================
void BGSetPalMap(u8 No,u32 size, u16 pal)
{
	u16 *BGMap = MEM_BG_MAP(No);
	u32 i;

	for(i=0; i<size/2; i++)
	{
		*BGMap++ |= pal<<12;
	}
}
//========================================================================
void BGSetPal16(u16 No, u16 *pal)
{
	u16 *BGPal = MEM_BG_PAL;
	u16 i;

	BGPal += 16*No;

	for(i=0; i<16; i++)
	{
		*BGPal++ = *pal++;
	}
}
//========================================================================
void BGSetPal256(u16 *pal)
{
	u16 *BGPal = MEM_BG_PAL;
	u16 i;
	
	for(i=0; i<256; i++)
	{
		*BGPal++ = *pal++;
	}
}
//========================================================================
void BGDrawStr(u8 BG,u16 x,u16 y,u8* str)
{
	u16 c;
	u16 i=0,dx=0;

	while(str[i]!=0x00 && str[i]!=0x0a)
	{
		if(str[i]<=0x7f)	//1bytechar
		{
			BGDrawChar(BG,x+dx,y,0,(u16)str[i]);
			dx++; i++;
		}
		else				//2bytechar
		{
			switch(str[i])
			{
				case 0x82: c = str[i+1]-0x90+192; break;
				case 0x83: c = str[i+1]-0x40+304; break;
				default  : c = str[i+1];          break;
			}
			BGDrawChar(BG,x+dx,y,0,c);
			dx++; i+=2;
		}
	}
}
//========================================================================
void BGDrawChar(u8 No, u16 x,u16 y,u16 palno,u16 chr)
{
	u16 *BGMap = MEM_BG_MAP(No);
	BGMap += y*32+x;
	*BGMap = chr+(palno<<12);
}
//========================================================================
u16  BGGetChar(u8 No, u8 x, u8 y)
{
	u16 *BGMap = MEM_BG_MAP(No);
	BGMap += y*32+x;
	return *BGMap;
}
//========================================================================
void BGSetChar(u8 No, u8 x, u8 y, u16 chr)
{
	u16 *BGMap = MEM_BG_MAP(No);
	BGMap += y*32+x;
	*BGMap = chr;
}
//========================================================================
//BG 256*256ΉĂȂ
void BGDrawCls(u8 BG, u16 chr)
{
	u16 *BGMap = MEM_BG_MAP(BG);
	u16 i;

	for(i=0; i<32*32; i++)
	{
		*(BGMap+i) = chr;
	}
}
//========================================================================
//BG 256*256ΉĂȂ
void BGDrawBox(u8 BG, u16 x, u16 y, u16 stopX, u16 stopY, u16 chr)
{
	u16 *BGMap = MEM_BG_MAP(BG);
	int j,k;

	BGMap += x * 0x01;	//x
	BGMap += y * 0x20;	//y

	for(j=0; j<stopY; j++)
	{
		for(k=0; k<stopX; k++)
		{
			*(BGMap+k) = chr;
		}
		BGMap += 0x20;
	}
}
//========================================================================
void BGDrawOrder(u8 BG)
{
	s16 x, y;
	u16 cnt = 0;

	for(y=0; y<20; y++)
	{
		for(x=0; x<30; x++)
		{
			BGSetChar(BG, x, y, cnt++);
		}
	}
}
//========================================================================
void SpriteInit()
{
	u8 i;
	for(i=0; i<MAX_SPRITE; i++)
	{
		SpriteMove(i, 240, 160);
	}
}
//========================================================================
void SpriteMove(u16 num, s16 x, s16 y)
{
	OAMEntry* sp = (OAMEntry*)OAMmem;
	sp += num;

	if(x<0) x += 512;
	if(y<0) y += 256;
	sp->attribute1 &= 0xfe00;
	sp->attribute0 &= 0xff00;
	sp->attribute1 |= (x & 0x01ff);
	sp->attribute0 |= (y & 0x00ff);
}
//========================================================================
void SpriteSetPal16No(u16 num, u16 palno)
{
	OAMEntry* sp = (OAMEntry*)OAMmem;
	sp += num;
	palno &= 0x000f;
	sp->attribute2 &= 0x0fff;
	sp->attribute2 |= (palno<<12);
}
//========================================================================
//	XvCǧ`ݒ
//	spEEE ύXXvCgw
//	sizeEE SP_SIZE_8,SP_SIZE_16,SP_SIZE_32,SP_SIZE_64
//	formEE SP_SQUARE,SP_TALL,SP_WIDE
//	colorEESP_COLOR_16, SP_COLOR_256
void SpriteSetSize(u16 num,u16 size,u16 form,u16 color)
{
	OAMEntry* sp = (OAMEntry*)OAMmem;
	sp += num;
	sp->attribute0 &= 0x1FFF;
	sp->attribute1 &= 0x3FFF;
	sp->attribute0 |= color | form | (160);
	sp->attribute1 |= size  | (240);
}
//========================================================================
void SpriteSetChr(u16 num, u16 ch)
{
	OAMEntry* sp = (OAMEntry*)OAMmem;
	sp += num;
	sp->attribute2 &= 0xFC00;
	sp->attribute2 |= ch;
}
//========================================================================
void SpriteSetPriority(u16 num, u16 priority)
{
	OAMEntry* sp = (OAMEntry*)OAMmem;
	sp += num;
	sp->attribute2 &= 0xF3FF;
	sp->attribute2 |= priority<<10;
}
//========================================================================
void SpriteSetVertical(u16 num, u8 flag)
{
	OAMEntry* sp = (OAMEntry*)OAMmem;
	sp += num;
	sp->attribute1 &= 0xEFFF;
	sp->attribute1 |= flag<<12;
}
//========================================================================
void SpriteSetData(u16 *img,u32 size) //size̓oCgP
{
	u16 *OAMDataAdr = OAMdata;
	u32 i;
	for(i=0; i<size/2; i++)
	{
		*OAMDataAdr++ = *img++;
	}
}
//========================================================================
void SpriteSetPal16(u16 *pal)
{
	u16* OBJPaleAdr = OBJpal;
	u16 i;
	for(i=0; i<16; i++)
	{
		*OBJPaleAdr++ = *pal++;
	}
}
//========================================================================
void SpriteSetPal256(u16 *pal)
{
	u16* OBJPaleAdr = OBJpal;
	u16 i;
	for(i=0; i<256; i++)
	{
		*OBJPaleAdr++ = *pal++;
	}
}
//========================================================================
char *bprintf(char *p, char *fmt, ...)
{
	va_list marker;
	va_start(marker, fmt);
	vsprintf(p, fmt, marker);
	va_end(marker);
	return p;
}
//========================================================================
int cprintf(char *fmt, ...)
{
	char s[80];
	va_list marker;
	va_start(marker, fmt);
	int r = vsprintf(s, fmt, marker);
	va_end(marker);
	cprint((u8*)s);
	return r;
}
//========================================================================
//VBÃfobNlogo͂
void cprint(u8* sz)
{
	const u32 port=0xc0ded00d;
	asm volatile(
		"mov r2, %0\n"
		"mov r0, %1\n"
		"mov r1, #0\n"
		"and r0, r0, r0\n"
		:
		:
		"r" (sz) ,"r" (port):
		"r0", "r1", "r2");
}
//========================================================================
  //܂̂ƂPCM(8bit Wav)t
void SoundInit()
{
	SOUNDCNT_X  = ENABLE_SOUND_MASTER;
	SOUNDCNT_L  = 0;
	SOUNDCNT_H  = 0;
	SOUNDCNT_H |= (DSOUND_A_TIMER0 | ENABLE_DSOUND_A_RIGHT | ENABLE_DSOUND_A_LEFT | DSOUND_A_FULL_OUTPUT);
	SOUNDCNT_H |= (RESET_DSOUND_A);
}
//========================================================================
void SoundSetData(s8 *offset, u32 size)
{
	WAVE.Offset = offset;
	WAVE.size   = size;
}
//========================================================================
void SoundStart()
{
	REG_TM0CNT   &= ~TM_ENABLE;
	REG_DM1CNT_H &= ~DMA_TRANSFER_ON;

	REG_DM1CNT_H = 0;
	REG_DM1SAD   = (u32)(WAVE.size);
	REG_DM1DAD   = (u32)(FIFO_A);
	REG_DM1CNT_H = (DMA_SIZE_32 | DMA_TRANSFER_OFF | DMA_TIMING_FIFO | 
	                DMA_REPEAT_ON | DMA_SAD_INC | DMA_DAD_FIX | DMA_INTR_ON);
	REG_TM0D     = (u16)(-1520);
	REG_TM0CNT   = TM_ENABLE | TM_FREQ_PER_1;

	REG_DM1CNT_H |= DMA_TRANSFER_ON;
}
//========================================================================
void *_malloc(u32 size)
{
	void *ptr;
	ptr=malloc(size);

	if(ptr!=NULL)
	{
		cprintf("_malloc ptr=0x%x  size=0x%x(%d) \n",(u32)ptr, size, size);
		return(ptr);
	}

	REG_IME = IRQ_MASTER_OFF;
	cprintf("vIG[:EWRAMȂȂ܂B\n");
	cprintf("_malloc ptr=0x%x \n",(u32)ptr);
	for(;;){}
}
//========================================================================
void _mfree(void *ptr)
{
	cprintf("_mfree ptr=0x%x\n",(u32)ptr);
	free(ptr);
}
//========================================================================
//ExtendTinyFileSystem
u8* EtfsOffset(u32 data_offset)
{
	u8* ewram = ADR_EWRAM;
	ewram += data_offset;
	cprintf("EtfsOffset ptr=0x%x;\n",(u32)ewram);
	return(ewram);
}
//========================================================================
u8 EtfsReadu8(u32 data_offset)
{
	u8* ewram = ADR_EWRAM;
	ewram += data_offset;
	return(*ewram);
}
//========================================================================
u16 EtfsReadu16(u32 data_offset)
{
	u16* ewram = (u16*)ADR_EWRAM;
	ewram += data_offset;
	return(*ewram);
}
//========================================================================
u32 EtfsReadu32(u32 data_offset)
{
	u32* ewram = (u32*)ADR_EWRAM;
	ewram += data_offset;
	return(*ewram);
}
//========================================================================
u8 *EtfsfopenLZ77(u32 offset, u32 size)
{
	u8 *buf;
	buf = _malloc(size);	//size͉𓀃TCY
	LZ77UnCompWram((void*)EtfsOffset(offset), (void*)buf);	//BiosCall
	return(buf);
}
//========================================================================
/*u8 *Etfsfopen(u32 offset, u32 size)
{
	u8 *buf;
	buf = _malloc(size);

	return(NULL);
}*/
//========================================================================
void SeekSetChr(u8* pOffset, u32 size, u8 chr)
{
	SEEK.Now   = pOffset;
	SEEK.Start = pOffset;
	SEEK.End   = pOffset + size;
	SEEK.Chr   = chr;
}
//========================================================================
u8* SeekNext()
{
	while(SEEK.Now < SEEK.End)
	{
		SEEK.Now++;
		if(*SEEK.Now == SEEK.Chr)
		{
			return SEEK.Now;
		}
	}
	return 0;
}
//========================================================================
//TODO: PrevpHasMoreĂȂ
u8* SeekPrev()
{
	while(SEEK.Now > SEEK.Start)
	{
		SEEK.Now--;
		if(*SEEK.Now == SEEK.Chr)
		{
			return SEEK.Now;
		}
	}
	return NULL;
}
//========================================================================
u8  SeekHasMore()
{
	u8* Now = SEEK.Now;
	while(Now < SEEK.End)
	{
		Now++;
		if(*Now == SEEK.Chr)
		{
			return TRUE;
		}
	}
	return FALSE;
}
//========================================================================
u8* SeekGetNow()
{
	return SEEK.Now;
}
//========================================================================
// 1bittHgf[^16hbg\
void Mode3DrawSJISBlock(u8 font,int x,int y,u16 color)
{
	u16* screen = VideoBuffer;
	screen[(y*240)+(x+0)] = (font & 0x80) ? color : screen[(y*240)+(x+0)];
	screen[(y*240)+(x+1)] = (font & 0x40) ? color : screen[(y*240)+(x+1)];
	screen[(y*240)+(x+2)] = (font & 0x20) ? color : screen[(y*240)+(x+2)];
	screen[(y*240)+(x+3)] = (font & 0x10) ? color : screen[(y*240)+(x+3)];
	screen[(y*240)+(x+4)] = (font & 0x08) ? color : screen[(y*240)+(x+4)];
	screen[(y*240)+(x+5)] = (font & 0x04) ? color : screen[(y*240)+(x+5)];
	screen[(y*240)+(x+6)] = (font & 0x02) ? color : screen[(y*240)+(x+6)];
	screen[(y*240)+(x+7)] = (font & 0x01) ? color : screen[(y*240)+(x+7)];
}
//========================================================================
// tHgf[^̐擪numberԖڂ̂P\
void Mode3DrawSJISAscii(u8 x, u8 y, u16 num, u16 col)
{
//	int index = (num/32*64 + num%32);

//	Set1blocMode3DrawSJISAsciik1bit(screen,(u8*)&HANKAKU_FONT[index*8        ],x1  ,y1  ,color);
//	Set1block1bit(screen,(u8*)&HANKAKU_FONT[index*8   +32*8],x1  ,y1+8,color);
}
//========================================================================
// tHgf[^̐擪numԖڂ̂P\
void Mode3DrawSJISChar(u8 x, u8 y, u16 num, u16 col)
{
	u32 index = num * FONTSJIS.size;
	u8  i,j;

	for(j=0; j<FONTSJIS.cy; j++)
	{
		for(i=0; i<FONTSJIS.cx; i++)
		{
			Mode3DrawSJISBlock(FONTSJIS.Offset[index++],x+i*8, y+j, col);
		}
	}
}
//========================================================================
// Spꕶ\
void Mode3DrawSJIS(u8 x, u8 y, u8* chr, u16 col)
{
	u16 code = (chr[1]*256) + chr[0]; //t
	u16 i;

	for(i=0; i<FONTSJIS.cnt; i++)
	{
		if(code == FONTSJIS.FontSheet[i])
		{
			//ꏑ
			i -= (i/0x21);		//'\n'
			Mode3DrawSJISChar(x,y,i,col);
		}
	}
}
//========================================================================
void Mode3DrawSJISStr(u8 x,u8 y,u8 *text,u16 col)
{
	u16 i=0;
	while( text[i] != '\0' )
	{
		if( text[i] >= 0x80 )	//Sp
		{
			Mode3DrawSJIS(x,y,&text[i],col);
			i+=2; x+=14;
		}
		else
		{						//p
			Mode3DrawSJISAscii(x,y,text[i],col);
			i+=1; x+=1;
		}
	}
}
//========================================================================
void WaitKey(u16 waitkey)
{
	REG_TM2CNT = 0;
	REG_TM2CNT = TM_FREQ_PER_1 | TM_ENABLE;

	for(;;)
	{
		if( !(*KEYS & waitkey) )
		{
			break;
		}
	}

	srand(REG_TM2D);

	REG_TM2CNT = 0;
	REG_TM2D   = 0;

	while(*KEYS != 0x3ff){}		//͂ɂȂ܂ő҂
}
//========================================================================
void WaitKey2(u16 waitkey)
{
	WaitKey(waitkey);
	WaitKey(waitkey);
}
//========================================================================
u16 KeyInput()
{
	KEY.old = KEY.cur;
	KEY.cur = ~(*KEYS);
	KEY.trg = (KEY.trg ^ KEY.cur) & (~KEY.old);
	KEY.off = (KEY.off) ^ (~KEY.cur & KEY.old);

	return KEY.cur;
}
//========================================================================
void KeyInit()
{
	KEY.cur = 0;	// Jg
	KEY.old = 0;	// O
	KEY.trg = 0;	// gK
	KEY.off = 0;	// ItgK

	Timer.wait      = 0;	//͎̃^C}
	Timer.repeat    = 0;
	Timer.clockTick = 0;
}
//========================================================================
u8* SRAMGetOffset(u32 offset)
{
	u8* sram = (u8*)ADR_SRAM;
	sram += offset;
	return sram;
}
//========================================================================
u8 SRAMRead8(u32 offset)
{
	u8* sram = (u8*)ADR_SRAM;
	sram += offset;
	return (u8)sram[0];
}
//========================================================================
u16 SRAMRead16(u32 offset)
{
	u8* sram = (u8*)ADR_SRAM;
	sram += offset;
	return (u16)sram[0] | (u16)sram[1]<<8;
}
//========================================================================
u32 SRAMRead32(u32 offset)
{
	u8* sram = (u8*)ADR_SRAM;
	sram += offset;
	return (u32)sram[0] | (u32)sram[1]<<8 | (u32)sram[2]<<16 | (u32)sram[3]<<24;
}
//========================================================================
void SRAMWrite8(u32 offset, u8 data)
{
	u8* sram = (u8*)ADR_SRAM;
	sram += offset;
	*sram = (u8)(data & 0xff);
}
//========================================================================
void SRAMWrite16(u32 offset, u16 data)
{
	u8* sram = (u8*)ADR_SRAM;
	sram += offset;
	*sram++ = (u8)((data   ) & 0x00ff);
	*sram   = (u8)((data>>8) & 0x00ff);
}
//========================================================================
void SRAMWrite32(u32 offset, u32 data)
{
	u8* sram = (u8*)ADR_SRAM;
	sram += offset;
	*sram++ = (u8)((data    ) & 0x000000ff);
	*sram++ = (u8)((data>>8 ) & 0x000000ff);
	*sram++ = (u8)((data>>16) & 0x000000ff);
	*sram++ = (u8)((data>>24) & 0x000000ff);
}
//========================================================================
//deg 0`255
s16 _Sin(u8 deg)
{
	return m_Sintbl[(deg) & 0xff];
}
//========================================================================
//deg 0`255
s16 _Cos(u8 deg)
{
	return m_Sintbl[((deg)+64) & 0xff];
}
//========================================================================
u8 _Angle(s16 dx, s16 dy)
{
	//̔߂
	s16 raito;
	if(dx == 0)
	{
		raito = 32767;
	}
	else
	{
		raito = (dy*256)/dx;
		if(raito < 0)
		{
			raito = -raito;
		}
	}

	//64T[`
	u8 angle;
	for(angle=0; angle<64; angle++)
	{
		if(raito < m_Atntbl[angle])
		{
			break;
		}
	}

	//𒲐
	if(dx < 0) angle = 64 - angle +  64;
	if(dy < 0) angle = 64 - angle + 192;

	return (angle & 0xff);
}
//========================================================================
void DMAChainInit()
{
	u8 i;
	for(i=0; i<MAX_DMA_CHAIN; i++)
	{
		DMA_CHAIN[i].src  = 0;
		DMA_CHAIN[i].dst  = 0;
		DMA_CHAIN[i].size = 0;
		DMA_CHAIN[i].mode = 0;
	}
}
//========================================================================
void DMAChainEnqueue(u32 src, u32 dst, u16 size, u16 mode)
{
}
//========================================================================
void DMAChainTransfer()
{
}
//========================================================================
u8 HitChkBox(u16 srcX, u16 srcY, u16 srcCx, u16 srcCy, u16 dstX, u16 dstY, u16 dstCx,u16 dstCy)
{
	if(    (srcY < dstY + dstCy)
	    && (dstY < srcY + srcCy) )
	{
		if(    (srcX < dstX + dstCx)
		    && (dstX < srcX + srcCx) )
		{
			return TRUE;
		}
	}
	return FALSE;
}
//========================================================================
u8 HitChkCircle(u16 srcX, u16 srcY, u16 srcR, u16 dstX, u16 dstY, u16 dstR)
{
	s16 x = srcX - dstX;
	s16 y = srcY - dstY;
	s16 r = srcR + dstR;

//Mode3DrawCircle(srcX, srcY, srcR, RGB(31,31,31));
//Mode3DrawCircle(dstX, dstY, dstR, RGB(31,31,31));

	if (x*x + y*y < r*r)	//s^SX̒藝
	{
		return TRUE;
	}
	return FALSE;
}
//========================================================================
void Mode3DrawCircle(u16 srcX, u16 srcY, u16 radius, u16 color)
{
	u8  x = radius;
	u8  y = 0;
	s32 p = 3 - 2 * radius;

	while (x >= y)
	{
		Mode3DrawDot(srcX + x, srcY + y, color);
		Mode3DrawDot(srcX - x, srcY + y, color);
		Mode3DrawDot(srcX + x, srcY - y, color);
		Mode3DrawDot(srcX - x, srcY - y, color);
		Mode3DrawDot(srcX + y, srcY + x, color);
		Mode3DrawDot(srcX - y, srcY + x, color);
		Mode3DrawDot(srcX + y, srcY - x, color);
		Mode3DrawDot(srcX - y, srcY - x, color);

		if (p >= 0)
		{
			x--;
			p -= 4 * x;
		}

		y++;
		p += 4 * y + 2;
	}
}
//========================================================================
void Mode3DrawFillCircle(u16 srcX, u16 srcY, u16 radius, u16 color)
{
	u8  x = radius;
	u8  y = 0;
	s32 p = 3 - 2 * radius;

	while (x >= y)
	{
		Mode3DrawLine(srcX-x, srcY+y, srcX+x, srcY+y, color);
		Mode3DrawLine(srcX-x, srcY-y, srcX+x, srcY-y, color);
		Mode3DrawLine(srcX-y, srcY+x, srcX+y, srcY+x, color);
		Mode3DrawLine(srcX-y, srcY-x, srcX+y, srcY-x, color);

		if(p >= 0)
		{
			x--;
			p -= 4 * x;
		}

		y++;
		p += 4 * y + 2;
	}
}
//========================================================================
// gk]@\gݒɂ
//	numEEEύXXvCgw
//	no EEEgpXvCgRotatep[^̔ԍ
//	double_flagEE`̈Q{gpł悤ɂ邩
//		0Ȃʏ̕`̈ASP_SIZE_DOUBLEȂQ{
void SpriteSetRotateFlag(u16 num, u16 no, u16 double_flag)
{
	OAMEntry* sp = (OAMEntry*)OAMmem;
	sp += num;
	
	sp->attribute0 |= SP_ROTATION_FLAG;
	
	sp->attribute1 &= 0xC1FF;
	sp->attribute1 |= SP_ROTDATA(no);
	
	sp->attribute0 &= 0xFDFF;
	sp->attribute0 |= double_flag;
}
//========================================================================
void SpriteSetScale(u16 num, u16 xscale, u16 yscale)
{
	RotData* rot = (RotData*)OAMmem;
	rot += num;

	rot->pa = 256*100/xscale;
	rot->pb = 0*100/xscale;
	rot->pc = 0*100/yscale;
	rot->pd = 256*100/yscale;
}
//========================================================================
// g嗦%Ŏw
void SpriteSetRotate(u16 num, u8 angle, u16 scale)
{
	RotData* rot = (RotData*)OAMmem;
	rot += num;

	s16 CosAngle = (s16)( _Cos(angle) * (128 + scale) );
	s16 SinAngle = (s16)( _Sin(angle) * (128 + scale) );

	rot->pa =  CosAngle;
	rot->pb =  SinAngle;
	rot->pc = -SinAngle;
	rot->pd =  CosAngle;
}
//========================================================================
void DebugMessage(u8* msg)
{
	cprintf((char*)msg);
	for(;;){};
}
//========================================================================
void BGSetScroll(u8 No, u16 x, u16 y)
{

	switch(No)
	{
		case 0: REG_BG0HOFS = x; REG_BG0VOFS = y; break;
		case 1: REG_BG1HOFS = x; REG_BG1VOFS = y; break;
		case 2: REG_BG2HOFS = x; REG_BG2VOFS = y; break;
		case 3: REG_BG3HOFS = x; REG_BG3VOFS = y; break;
		default: DebugMessage((u8*)"Err: BGSetScroll No\n"); break;
	}
}
//========================================================================
bool KeyWaitTimer()
{
	if(Timer.clockTick - Timer.wait < TIMER_INTMS)
	{
		return TRUE;
	}

	Timer.wait = Timer.clockTick;
	return FALSE;
}
//========================================================================
//L[s[hi`^Oh~j
bool KeyRepeat()
{
	ST_INPUT* p = &Input;
	ST_TIMER* t = &Timer;

	p->keyNow   = KeyInput();

	if(p->keyNow && p->keyNow == p->keyOld)
	{
		if(t->clockTick - t->repeat < KEY_REPEAT_TIME)
		{
			return TRUE;
		}

		return FALSE;
	}

	p->keyOld = p->keyNow;
	t->repeat = t->clockTick;

	return FALSE;
}
//========================================================================
u8 KeyGet()
{
	ST_INPUT* p = &Input;
	return p->keyNow;
}
//========================================================================
void WaitMs(u32 wait)
{
	u32 t = Timer.clockTick;

	while(Timer.clockTick - t < wait)
	{
	}
}
